class Cache:
    _instance = None

    def __init__(self):
        if self._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.data = {}


    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    def read(self, key):
        return self.data.get(key)

    def write(self, key, val):
        self.data[key] = val

    def delete(self, key):
        del self.data[key]



import base64
import json
import requests
import pdb

class AgentApi:
    STATUSES_ENDPOINT = '/api/v1/admin/statuses'

    def __init__(self, service_url):
        self.service_url = service_url

    @property
    def default_headers(self):
        return {
            'X-Do-Proxy': '1',
        }

    def update_status(self, status_id, project_key):
        project_data = self.decode_project_key(project_key)
        url = f"{self.service_url}{self.STATUSES_ENDPOINT}/{status_id}"
        return requests.put(url, data=str(project_data.get('id')), headers=self.default_headers)

    @staticmethod
    def decode_project_key(key):
        # TODO: add specific error catching
        try:
            key = base64.b64decode(key)
        except:
            return {}

        # TODO: add specific error catching
        try:
            return json.loads(key)
        except:
            return {}



LOG_LEVEL = 'STOOBLY_AGENT_LOG_LEVEL'
AGENT_ACTIVE_MODE = 'STOOBLY_AGENT_ACTIVE_MODE'
AGENT_ENABLED = 'STOOBLY_AGENT_ENABLED'
AGENT_IS_HEADLESS = 'STOOBLY_AGENT_IS_HEADLESS'
AGENT_INCLUDE_PATTERNS = 'STOOBLY_AGENT_INCLUDE_PATTERNS'
AGENT_EXCLUDE_PATTERNS = 'STOOBLY_AGENT_EXCLUDE_PATTERNS'
AGENT_POLICY = 'STOOBLY_AGENT_POLICY'
AGENT_SERVICE_URL = 'STOOBLY_AGENT_SERVICE_URL'
AGENT_PROJECT_KEY = 'STOOBLY_AGENT_PROJECT_KEY'
AGENT_PROXY_URL = 'STOOBLY_AGENT_PROXY_URL'
AGENT_SCENARIO_KEY = 'STOOBLY_AGENT_SCENARIO_KEY'
AGENT_URL = 'STOOBLY_AGENT_URL'
API_URL = 'STOOBLY_API_URL'
API_KEY = 'STOOBLY_API_KEY'



import os
import time

from pathlib import Path
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

def current_working_directory():
    return os.path.dirname(os.path.realpath(__file__))

def trigger_reload(event):
    print('Hot reloading...')

    cwd = current_working_directory()
    Path(os.path.join(current_working_directory(), 'record.py')).touch()

cwd = current_working_directory()
lib_path = os.path.join(cwd, 'lib')

patterns = "*"
ignore_patterns = ""
ignore_directories = False
case_sensitive = True
event_handler = PatternMatchingEventHandler(patterns, ignore_patterns, ignore_directories, case_sensitive)

event_handler.on_modified = trigger_reload

observer = Observer()
observer.schedule(event_handler, lib_path, recursive=True)
observer.start()

try:
    while True:
        time.sleep(1)
finally:
    observer.stop()
    observer.join()



import re
import time
import pdb

class ResponseString:
    ENCODING = 'utf-8'
    RESPONSE_TYPE = 2
    CLRF = b"\r\n"

    def __init__(self, response, request_id):
        self.response = response
        self.lines = []
        self.latency = 0
        self.request_id = request_id

        self.response_line()
        self.headers()
        self.body()

    ###
    #
    # 1 - response type
    # 2 - request id
    # 3 - timestamp in nano seconds
    # 4 - response latency in nano seconds
    #
    def control(self):
        current_time = self.current_time()

        line = "{} {} {} {}".format(self.RESPONSE_TYPE, self.request_id, current_time, self.latency)
        self.lines.insert(0, line.encode(self.ENCODING))

    def response_line(self):
        line = "HTTP/1.1 {}".format(self.response.code)
        self.lines.append(line.encode(self.ENCODING))

    def headers(self):
        headers = self.response.headers

        for name, val in headers.items():
            line = "{}: {}".format(self.to_header_case(name), val)
            self.lines.append(line.encode(self.ENCODING))

    def body(self):
        body = self.response.body

        if not body:
            self.lines.append(self.CLRF)
        elif isinstance(body, bytes):
            self.lines.append(self.CLRF + body)
        else:
            raise Exception('Unsupported body type')

    def with_latency(self, latency):
        self.latency = latency

    def get(self):
        self.control()
        return self.CLRF.join(self.lines)

    def to_header_case(self, header):
        toks = re.split('_|-', header)
        if len(toks) == 1:
            return header

        for index, tok in enumerate(toks):
            toks[index] = tok.lower().capitalize()
        return "-".join(toks)

    def current_time(self):
        now = time.time()
        current_time = round(now * (pow(10, 6)))
        return current_time



import errno
import importlib
import json
import os
import pdb
import re
import tempfile
import time
import threading

from mitmproxy import http
from urllib.parse import urlparse

import lib

from lib.agent_api import AgentApi
from lib.hashed_request_decorator import HashedRequestDecorator
from lib.joined_request import JoinedRequest
from lib.logger import Logger
from lib.mitmproxy_request_adapter import MitmproxyRequestAdapter
from lib.mitmproxy_response_adapter import MitmproxyResponseAdapter
from lib.proxy_request import ProxyRequest
from lib.settings import Settings
from lib.stoobly_api import StooblyApi

# mitmproxy only hot reloads the main script, manually hot reload lib
importlib.reload(lib.hashed_request_decorator)
importlib.reload(lib.joined_request)
importlib.reload(lib.logger)
importlib.reload(lib.mitmproxy_request_adapter)
importlib.reload(lib.mitmproxy_response_adapter)
importlib.reload(lib.proxy_request)
importlib.reload(lib.settings)
importlib.reload(lib.stoobly_api)

# Disable proxy settings in urllib
os.environ['no_proxy'] = '*'

# Observe config for changes
Settings.instance().observe_config()

LOG_ID = 'Record'

AGENT_STATUSES = {
    'REQUESTS_MODIFIED': 'requests-modified'
}

MOCK_POLICY = {
    'ALL': 'all',
    'NONE': 'none',
    'FOUND': 'found',
}

RECORD_POLICY = {
    'NONE': 'none',
    'ALL': 'all',
    'NOT_FOUND': 'not_found',
}

MODE = {
    'MOCK': 'mock',
    'NONE': 'none',
    'RECORD': 'record',
}

CUSTOM_RESPONSE_CODES = {
    'NOT_FOUND': 499,
    'IGNORE_COMPONENTS': 498,
}

CUSTOM_HEADERS = {
    'MOCK_POLICY': 'X-Mock-Policy',
    'DO_PROXY': 'X-Do-Proxy',
    'PROXY_MODE': 'X-Proxy-Mode',
    'RECORD_POLICY': 'X-Record-Policy',
    'RESPONSE_LATENCY': 'X-Response-Latency',
    'SERVICE_URL': 'X-Service-Url',
}

NAMESPACE_FOLDER = 'stoobly'

def request(flow):
    request = flow.request

    __disable_web_cache(request)

    settings = Settings.instance()
    mode = __get_proxy_mode(request.headers, settings)

    Logger.instance().debug(f"{LOG_ID}:ProxyMode: {mode}")

    if mode == MODE['NONE']:
        pass
    elif mode == MODE['RECORD']:
        __handle_record(request, settings)
    elif mode == MODE['MOCK']:
        __handle_mock(flow, settings)
    else:
        return __bad_request(
            flow,
            "Valid env MODE: %s, %s, Got: %s" % (MODE['RECORD'], MODE['MOCK'], mode)
        )

def response(flow):
    settings = Settings.instance()
    request = flow.request

    mode = __get_proxy_mode(request.headers, settings)

    if mode != MODE['RECORD']:
        return False

    __disable_transfer_encoding(flow.response)

    active_mode_settings = settings.active_mode_settings

    api = StooblyApi(
      settings.api_url, settings.api_key
    )

    if active_mode_settings.get('enabled') and __allowed_request(active_mode_settings, request):
        upload_policy = __get_record_policy(request.headers, active_mode_settings)
    else:
        # If the request path does not match accepted paths, do not record
        upload_policy = RECORD_POLICY['NONE']

    Logger.instance().debug(f"{LOG_ID}:UploadPolicy: {upload_policy}")

    if upload_policy == RECORD_POLICY['ALL']:
        thread = threading.Thread(target=__upload_request, args=(flow, api, settings))
        thread.start()
        #__upload_request(flow, api, settings)
    elif upload_policy == RECORD_POLICY['NOT_FOUND']:
        res = __eval_request(request, api, active_mode_settings)

        if res.status_code == CUSTOM_RESPONSE_CODES['NOT_FOUND']:
            thread = threading.Thread(target=__upload_request, args=(flow, api, settings))
            thread.start()
            #__upload_request(flow, api, settings)
    elif upload_policy == RECORD_POLICY['NONE']:
        pass
    else:
        return __bad_request(
            flow,
            "Valid env RECORD_POLICY: %s, %s, %s, Got: %s" %
            [RECORD_POLICY['ALL'], RECORD_POLICY['NOT_FOUND'], RECORD_POLICY['NONE'], upload_policy]
        )

### PRIVATE

###
#
# @param request [mitmproxy.net.http.request.Request]
# @param settings [Dict]
#
def __handle_mock(flow, settings):
    start_time = time.time()

    request = flow.request
    active_mode_settings = settings.active_mode_settings
    service_url = __get_service_url(request, active_mode_settings)

    api = StooblyApi(
      settings.api_url, settings.api_key
    )

    if active_mode_settings.get('enabled') and __allowed_request(active_mode_settings, request):
        mock_policy = __get_mock_policy(request.headers, active_mode_settings)
    else:
        # If the request path does not match accepted paths, do not mock
        mock_policy = MOCK_POLICY['NONE']

    if mock_policy == MOCK_POLICY['NONE']:
        return __reverse_proxy(request, service_url, {})
    elif mock_policy == MOCK_POLICY['ALL']:
        res = __eval_request(request, api, active_mode_settings)

        if res.status_code == CUSTOM_RESPONSE_CODES['IGNORE_COMPONENTS']:
            res = __eval_request(request, api, active_mode_settings, res.content)

            __simulate_latency(res.headers.get(CUSTOM_HEADERS['RESPONSE_LATENCY']), start_time)
    elif mock_policy == MOCK_POLICY['FOUND']:
        res = __eval_request(request, api, active_mode_settings)

        if res.status_code == CUSTOM_RESPONSE_CODES['IGNORE_COMPONENTS']:
            res = __eval_request(request, api, active_mode_settings, res.content)

        if res.status_code == CUSTOM_RESPONSE_CODES['NOT_FOUND']:
            return __reverse_proxy(request, service_url, {})
        else:
            __simulate_latency(res.headers.get(CUSTOM_HEADERS['RESPONSE_LATENCY']), start_time)
    else:
        return __bad_request(
            flow,
            "Valid env MOCK_POLICY: %s, %s, %s, Got: %s" %
            [MOCK_POLICY['ALL'], MOCK_POLICY['FOUND'], MOCK_POLICY['NONE'], mock_policy]
        )

    return __pass_on(flow, res)

def __handle_record(request, settings):
    active_mode_settings = settings.active_mode_settings

    service_url = __get_service_url(request, active_mode_settings)

    #
    # Try forwarding the request to the service specified by Settings.service_url
    #
    if not service_url:
        raise Exception('config service_url is not set')

    __reverse_proxy(request, service_url, {})

### API Access

def __reverse_proxy(request, service_url, options = {}):
    Logger.instance().debug(f"{LOG_ID}:ReverseProxy:ServiceUrl: {service_url}")

    uri = urlparse(service_url)

    request.scheme = uri.scheme
    request.host = uri.hostname
    if uri.port:
        request.port = uri.port

###
#
# Upon receiving a response, create the request in API for future use
#
# @param api [StooblyApi]
# @param settings [Settings.mode.mock | Settings.mode.record]
# @param res [Net::HTTP::Response]
#
def __upload_request(flow, api, settings):
    active_mode_settings = settings.active_mode_settings
    service_url = __get_service_url(flow.request, active_mode_settings)
    request = MitmproxyRequestAdapter(flow.request)
    proxy_request = ProxyRequest(request, service_url)
    response = MitmproxyResponseAdapter(flow.response)

    joined_request = JoinedRequest(proxy_request).with_response(response)

    Logger.instance().info(f"Uploading {proxy_request.url()}")

    raw_requests = joined_request.build()

    res = api.request_create(
        active_mode_settings.get('project_key'),
        raw_requests,
        {
            'importer': 'gor',
            'scenario_key': active_mode_settings.get('scenario_key'),
        }

    )

    Logger.instance().debug(f"{LOG_ID}:UploadRequest:StatusCode:{res.status_code}")

    # Write the request to a file to help debug
    if Settings.instance().is_debug():
        # Build file path, replace slashes with underscores
        request_path = request.path.replace('/', '_')
        timestamp = str(int(time.time() * 1000))
        file_path = os.path.join(tempfile.gettempdir(), NAMESPACE_FOLDER, request_path, timestamp)

        if not os.path.exists(os.path.dirname(file_path)):
            try:
                os.makedirs(os.path.dirname(file_path))
            except OSError as err: # Guard against race condition
                if err.errno != errno.EEXIST:
                    raise err

        Logger.instance().debug(f"{LOG_ID}: Writing request to {file_path}")

        with open(file_path, 'wb') as f:
            f.write(raw_requests)

    if not Settings.instance().is_headless() and res.status_code == 201:
        agent_url = settings.agent_url

        if not agent_url:
            Logger.instance().warn('Settings.agent_url not configured')
        else:
            api = AgentApi(agent_url)
            api.update_status(AGENT_STATUSES['REQUESTS_MODIFIED'], active_mode_settings.get('project_key'))

    return res


###
#
# @param api [StooblyApi]
# @param settings [Settings.mode.mock | Settings.mode.record]
# @param ignored_components_json [String] JSON string
#
def __eval_request(request, api, settings, ignored_components_json = None):
    ignored_components = []

    if ignored_components_json:
        try:
            ignored_components = json.loads(ignored_components_json)
        except:
            pass

    query_params = __build_query_params(request, ignored_components)
    query_params['scenario_key'] = settings.get('scenario_key')

    return api.request_response(
        settings.get('project_key'), query_params
    )

### Helpers

###
#
# Return response headers, body, and status code
#
def __pass_on(flow, res):
    # Ideally we just return the HTTPResponse object from res.raw
    # See Issue #11
    #flow.response = res.raw

    headers = {}
    for key, value in res.headers.items():
        headers[key.capitalize()] = value

    # Without specifying a length to read, requests will compare content length
    # with Content-Length header. If the content is gzipped, an IncompleteRead error will be thrown
    #content = res.raw.read(res.raw.length_remaining)

    content = res.content

    flow.response = http.HTTPResponse.make(
        res.status_code, content, headers,
    )

def __bad_request(flow, message):
    flow.response = http.HTTPResponse.make(
        400,  # (optional) status code
        message,
        {'Content-Type': 'text/plain'}  # (optional) headers
    )

    return False

def __allowed_request(active_mode_settings, request):
    exclude_patterns = active_mode_settings.get('exclude_patterns')

    if exclude_patterns:
        if __exclude(request,  exclude_patterns):
            return False

    # If an include pattern is set, then that means only requests
    # matching these pattern(s) are allowed
    include_patterns = active_mode_settings.get('include_patterns')
    if include_patterns:
        if not __include(request, include_patterns):
            return False

    # If there are no exclude or include patterns, request is allowed
    return True

###
#
# @param patterns [Array<string>]
#
def __include(request, patterns):
    if not patterns:
        return True

    if len(patterns) == 0:
        return True

    for pattern in patterns:
        if re.match(pattern, request.url):
            return True

    return False

def __exclude(request, patterns):
    if not patterns:
        return False

    for pattern in patterns:
        if re.match(pattern, request.url):
            return True

    return False

###
#
# Formats request into parameters expected by stoobly api
#
# @param request [lib.mitmproxy_request_adapter.MitmproxyRequestAdapter]
# @param ignored_components [Array<Hash>]
#
# @return [Hash] query parameters to pass to stoobly api
#
def __build_query_params(request, ignored_components = []):
    request = MitmproxyRequestAdapter(request)
    hashed_request = HashedRequestDecorator(request).with_ignored_components(ignored_components)

    headers_hash = hashed_request.headers_hash()
    query_params_hash = hashed_request.query_params_hash()
    body_params_hash = hashed_request.body_params_hash()
    body_text_hash = hashed_request.body_text_hash()

    query_params = {}
    query_params['host'] = request.host
    query_params['path'] = request.path
    query_params['port'] = request.port
    query_params['method'] = request.method

    if len(headers_hash) > 0:
        query_params['headers_hash'] = headers_hash

    if len(query_params_hash) > 0:
        query_params['query_params_hash'] = query_params_hash

    # The problem here is that if a body_param_hash is not passed, then body_text_hash is checked
    # However, if body_params are ignored, then there may not be a body_param_hash
    #
    # If there are ignored body_params, this means there are body_params
    # Then we should use body_params_hash over body_text_hash
    if len(body_params_hash) > 0 or len(hashed_request.ignored_body_params) > 0:
        query_params['body_params_hash'] = body_params_hash

    if len(body_text_hash) > 0:
        query_params['body_text_hash'] = body_text_hash

    if len(ignored_components) > 0:
        query_params['retry'] = 1

    return query_params

###
#
# Try to simulate expected response latency
#
# wait_time (seconds) = expected_latency - estimated_rtt_network_latency - api_latency
#
# expected_latency = provided value
# estimated_rtt_network_latency = 15ms
# api_latency = current_time - start_time of this request
#
def __simulate_latency(expected_latency, start_time):
    if not expected_latency:
        return 0

    estimated_rtt_network_latency = 0.015 # seconds
    api_latency = (time.time() - start_time)
    expected_latency = float(expected_latency) / 1000

    wait_time = expected_latency - estimated_rtt_network_latency - api_latency

    Logger.instance().debug(f"{LOG_ID}:Expected latency: {expected_latency}")
    Logger.instance().debug(f"{LOG_ID}:API latency: {api_latency}")
    Logger.instance().debug(f"{LOG_ID}:Wait time: {wait_time}")

    if wait_time > 0:
        time.sleep(wait_time)

    return wait_time

### Setters

def __disable_transfer_encoding(response):
    if 'Transfer-Encoding' in response.headers:
        # Without deleting this header, causes caller to stall
        del response.headers['Transfer-Encoding']

def __disable_web_cache(request):
    request.headers['CACHE-CONTROL'] = 'no-cache'

    if 'IF-NONE-MATCH' in request.headers:
        del request.headers['IF-NONE-MATCH']

### Getters

def __get_proxy_mode(headers, settings):
    access_control_header =  'Access-Control-Request-Headers'
    do_proxy_header = CUSTOM_HEADERS['DO_PROXY']

    if access_control_header in headers and do_proxy_header.lower() in headers[access_control_header]:
        return MODE['NONE']
    elif do_proxy_header in headers:
        return MODE['NONE']
    elif CUSTOM_HEADERS['PROXY_MODE'] in headers:
        return headers[CUSTOM_HEADERS['PROXY_MODE']]
    else:
        return settings.active_mode

def __get_mock_policy(headers, settings):
    if CUSTOM_HEADERS['MOCK_POLICY'] in headers:
        return headers[CUSTOM_HEADERS['MOCK_POLICY']]
    else:
        return settings.get('policy')

def __get_record_policy(headers, settings):
    if CUSTOM_HEADERS['RECORD_POLICY'] in headers:
        return headers[CUSTOM_HEADERS['RECORD_POLICY']]
    else:
        return settings.get('policy')

def __get_service_url(request, settings):
    service_url = request.headers.get(CUSTOM_HEADERS['SERVICE_URL'])

    if service_url:
        return service_url
    else:
        if settings.get('service_url') and len(settings.get('service_url')) > 0:
            return settings.get('service_url')

        return self.__upstream_url(request)

def __upstream_url(request):
    return f"{request.scheme}://{request.host}:{request.port}"




import base64
import json
import requests
import urllib
import pdb

from .logger import Logger

class StooblyApi:
    LOG_ID = 'lib.stoobly_api'
    REQUESTS_ENDPOINT = '/requests'

    def __init__(self, service_url, api_key):
        self.service_url = service_url
        self.api_key = api_key

    @staticmethod
    def decode_project_key(key):
        # TODO: add specific error catching
        try:
            key = base64.b64decode(key)
        except:
            return {}

        # TODO: add specific error catching
        try:
            return json.loads(key)
        except:
            return {}

    @staticmethod
    def decode_scenario_key(key):
        try:
            key = base64.b64decode(key)
        except:
            return {}

        try:
            return json.loads(key)
        except:
            return {}

    @property
    def default_headers(self):
        return {
            'X-API-KEY': self.api_key,
            'X-Do-Proxy': '1',
        }

    def request_create(self, project_key, raw_requests, params):
        url = f"{self.service_url}{self.REQUESTS_ENDPOINT}"

        self.__parse_scenario_key(params)

        project_data = self.decode_project_key(project_key)

        body = {
            'project_id': project_data.get('id'),
            **params,
        }

        return requests.post(url, headers=self.default_headers, data=body, files={ 'requests': raw_requests })

    def request_response(self, project_key, query_params):
        url = f"{self.service_url}{self.REQUESTS_ENDPOINT}/response"

        self.__parse_scenario_key(query_params)

        project_data = self.decode_project_key(project_key)

        params = {
            'project_id': project_data.get('id'),
            **query_params,
        }

        Logger.instance().debug(f"{self.LOG_ID}.request_response:{url}?{urllib.parse.urlencode(params)}")

        return requests.get(
            url,
            allow_redirects=False,
            headers=self.default_headers,
            params=params,
            stream=True
        )

    def __parse_scenario_key(self, params):
        if not 'scenario_key' in params:
            return

        if len(params['scenario_key']) != 0:
            scenario_data = self.decode_scenario_key(params['scenario_key'])

            if 'id' in scenario_data:
                scenario_id = scenario_data['id']
                params['scenario_id'] = scenario_id

        del params['scenario_key']



import time

from .request_string import RequestString
from .response_string import ResponseString

class JoinedRequest:
    ENCODING = 'utf-8'
    REQUEST_DELIMITTER = '🐵🙈🙉'.encode(ENCODING)

    ###
    #
    # @params proxy_request [ProxyRequest]
    #
    def __init__(self, proxy_request):
        self.timestamp = time.time()
        self.proxy_request = proxy_request

        self.request_string = RequestString(proxy_request)

    def with_response(self, response):
        now = time.time()

        self.response_string = ResponseString(response, self.request_string.request_id)

        # milliseconds
        latency = round((now - self.timestamp) * (10 ** 6))
        self.response_string.with_latency(latency)

        return self

    def build(self):
        if not self.response_string:
            raise Exception('Missing response')

        request_string = self.request_string.get()
        response_string = self.response_string.get()

        joined = self.REQUEST_DELIMITTER.join([request_string, response_string])

        return joined



import os

class RootDir:
    HOME_DIR = os.path.expanduser('~')
    DIR_NAME = '.stoobly'

    _instance = None

    def __init__(self):
        if RootDir._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.root_dir = os.path.join(self.HOME_DIR, self.DIR_NAME)
            if not os.path.exists(self.root_dir):
                os.mkdir(self.root_dir)

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    @property
    def tmp_dir(self):
        return os.path.join(self.root_dir, 'tmp')




import json
import mimetypes
import os
import pathlib
import pdb
import re
import yaml

from http.server import HTTPServer, BaseHTTPRequestHandler
from mergedeep import merge
from urllib.parse import urlparse, parse_qs

from .app.configs_controller import ConfigsController
from .app.proxy_controller import ProxyController
from .app.statuses_controller import StatusesController
from .lib import headers

class SimpleHTTPRequestHandler(BaseHTTPRequestHandler):
    ROOT_DIR = os.path.dirname(os.path.realpath(__file__))

    CONFIGS_PATH = '/api/v1/admin/configs'
    PROXY_PATH = '/proxy'
    STATUSES_PATH = '/api/v1/admin/statuses'
    STATUS_PATH = re.compile(f"{STATUSES_PATH}/.*[^/]$")

    @property
    def public_dir(self):
        return os.path.join(self.ROOT_DIR, 'public')

    ### Method handlers

    def do_OPTIONS(self):
        self.render(
            plain = 'OK',
            status = 200
        )

    def do_GET(self):
        self.preprocess()

        if not self.route('GET'):
            path = os.path.join(self.public_dir, self.path[1:] if self.path != '/' else 'index.html')
            if not os.path.exists(path):
                path = os.path.join(self.public_dir, 'index.html')

            self.render(
                file = path,
                status = 200
            )

    def do_PUT(self):
        self.preprocess()

        if not self.route('PUT'):
            self.render(
                plain = 'NOT FOUND',
                status = 404
            )

    ### Helpers

    def enable_cors(self):
        self.send_header('Access-Control-Allow-Methods', 'GET, OPTIONS, POST, PATCH, PUT, DELETE')
        self.send_header('Access-Control-Allow-Origin', '*')

        allowed_headers = ', '.join([
            'Content-Type'.upper(),
            headers.DO_PROXY.upper(),
            headers.PROXY_HEADERS.upper(),
            headers.REQUEST_PATH.upper(),
            headers.SERVICE_URL.upper(),
        ])
        self.send_header('Access-Control-Allow-Headers', allowed_headers)

        self.send_header('Access-Control-Max-Age', '7200')

    def preprocess(self):
        self.uri = urlparse(self.path)

        self.fullpath = self.path
        self.path = self.uri.path
        self.params = {}
        self.parse_query_params()

    def route(self, method):
        for endpoint_handler in self.ROUTES[method]:
            path = endpoint_handler[0]

            matches = isinstance(path, str) and self.path == path
            matches = matches or not isinstance(path, str) and re.match(path, self.path)

            if matches:
                handler = endpoint_handler[1]
                handler(self)
                return True

        return False

    def parse_path_params(self, path_params_map):
        path = self.uri.path
        path_segments = path.split('/')

        if len(path_segments) == 0:
            return

        if path_segments[0] == '':
            path_segments.pop(0)

        for path_param in path_params_map:
            index = path_params_map[path_param]

            try:
                self.params[path_param] = path_segments[index]
            except:
                pass

    def parse_query_params(self):
        merge(self.params, parse_qs(self.uri.query))

    def parse_body(self):
        if not self.headers.get('Content-Length'):
            body = self.rfile.read()
        else:
            content_length = int(self.headers['Content-Length'])
            body = self.rfile.read(content_length)

        if not self.headers['Content-Type']:
            return body

        if self.headers['Content-Type'].lower() == 'application/json':
            return json.loads(body)
        else:
            return body

    def render(self, **kwargs):
        if kwargs.get('file') != None:
            self.render_file(kwargs)
        elif kwargs.get('json') != None:
            self.render_json(kwargs)
        elif kwargs.get('plain') != None:
            self.render_plain(kwargs)
        elif kwargs.get('data') != None:
            self.render_data(kwargs)

    def render_file(self, kwargs):
        if not 'headers' in kwargs:
            kwargs['headers'] = {}

        path = kwargs['file']

        if not os.path.exists(path):
            kwargs['status'] = 404

        mimetype = mimetypes.guess_type(path)[0]
        kwargs['headers']['Content-Type'] = mimetype or 'text/plain'

        fp = open(path, 'rb')
        kwargs['data'] = fp.read()
        self.render_data(kwargs)

    def render_json(self, kwargs):
        if not 'headers' in kwargs:
            kwargs['headers'] = {}

        kwargs['headers']['Content-Type'] = 'application/json'
        kwargs['data'] = json.dumps(kwargs['json'])
        self.render_data(kwargs)

    def render_plain(self, kwargs):
        if not 'headers' in kwargs:
            kwargs['headers'] = {}

        kwargs['headers']['Content-Type'] = 'text/plain'
        kwargs['data'] = kwargs['plain']
        self.render_data(kwargs)

    def render_data(self, kwargs):
        self.send_response(kwargs.get('status') or 200)

        body = b''
        if isinstance(kwargs['data'], str):
            body = kwargs['data'].encode()
        else:
            body = kwargs['data']

        # Send headers
        headers = self.filter_headers(kwargs.get('headers'), {
            'TRANSFER-ENCODING': 'CHUNKED',
        })
        headers['Content-Length'] = len(body)
        self.enable_cors()
        self.render_headers(headers)
        self.end_headers()

        # Send body
        self.wfile.write(body)

    def filter_headers(self, headers, blacklist = {}):
        if not headers:
            return {}

        new_blacklist = {}
        for key, val in blacklist.items():
            new_blacklist[key.upper()] = val

        new_headers = headers.copy()
        for key, val in headers.items():
            if key.upper() in new_blacklist:
                if not new_blacklist[key.upper()]:
                    del new_headers[key.upper()]
                elif val.upper() == new_blacklist[key.upper()]:
                    del new_headers[key.upper()]

        return new_headers

    def render_headers(self, headers):
        if headers:
            for key, val in headers.items():
                self.send_header(key, val)

    ### Routes

    ROUTES = {
        'GET': [
            [CONFIGS_PATH, ConfigsController.instance().get_configs],
            ['/'.join([CONFIGS_PATH, 'modes']), ConfigsController.instance().get_configs_modes],
            ['/'.join([CONFIGS_PATH, 'policies']), ConfigsController.instance().get_configs_policies],
            ['/'.join([PROXY_PATH, 'get']), ProxyController.instance().do_GET],
            [STATUS_PATH, StatusesController.instance().get_status],
        ],
        'PUT': [
            [CONFIGS_PATH, ConfigsController.instance().put_configs],
            ['/'.join([PROXY_PATH, 'put']), ProxyController.instance().do_PUT],
            [STATUS_PATH, StatusesController.instance().put_status],
        ]
    }

def run(host, port):
    httpd = HTTPServer((host, port), SimpleHTTPRequestHandler)
    httpd.serve_forever()



import logging
import os

from .env_vars import LOG_LEVEL

class Logger:
    _instance = None

    def __init__(self):
        raise RuntimeError('Call instance() instead')

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls.__new__(cls)

            log_level = os.getenv(LOG_LEVEL) or ''

            if log_level.lower() == 'debug':
                logging.basicConfig(level=logging.DEBUG)
            elif log_level.lower() == 'warning':
                logging.basicConfig(level=logging.WARNING)
            elif log_level.lower() == 'error':
                logging.basicConfig(level=logging.ERROR)
            else:
                logging.basicConfig(level=logging.INFO)

        return logging




import os
import tempfile

def run_command(shell, command, is_command):
    os.system(os.system(command) if is_command else os.system(f"{shell} {command}"))

def run_command_with_proxy_export(shell, command, is_command, proxy_url):
    # Create a temp file
    fd, tmp_path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, 'w+') as tmp:
            # Write commands to temp file
            export_command = f"export HTTP_PROXY={proxy_url}\nexport HTTPS_PROXY={proxy_url}\n"
            tmp.write(export_command)
            tmp.write(command if is_command else f"{shell} {command}")
            tmp.flush()

            # Run temp file
            os.system(f"{shell} {tmp_path}")
    finally:
        os.remove(tmp_path)



import json
import os
import yaml
import pdb

from shutil import copyfile
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

from . import env_vars
from .logger import Logger
from .root_dir import RootDir

class Settings:
    LOG_ID = 'lib.settings'
    FILE_NAME = 'settings.yml'

    _instance = None
    _agent_url = ''
    _proxy_url = ''
    _api_url = ''

    def __init__(self):
        if Settings._instance:
            raise RuntimeError('Call instance() instead')
        else:
            cwd = os.path.dirname(os.path.realpath(__file__))

            self.config_file_path = os.path.join(RootDir.instance().root_dir, self.FILE_NAME)

            # If the config does not exist, use template
            if not os.path.exists(self.config_file_path):
                self.config_template_file_path = os.path.join(cwd, '..', 'config', self.FILE_NAME)
                copyfile(self.config_template_file_path, self.config_file_path)

            self.__load_config()

    def observe_config(self):
        patterns = ['settings.yml']
        ignore_patterns = None
        ignore_directories = False
        case_sensitive = True
        event_handler = PatternMatchingEventHandler(patterns, ignore_patterns, ignore_directories, case_sensitive)
        event_handler.on_modified = self.reload_config

        observer = Observer()
        watch_dir = os.path.dirname(self.config_file_path)

        observer.schedule(event_handler, watch_dir)
        observer.start()

    ### Statuses

    # Headless means the agent is packaged without the frontend and
    # supports configuration with environment variables or the yaml file.
    def is_headless(self):
        if os.environ.get(env_vars.AGENT_IS_HEADLESS):
            return True

        return False

    def is_debug(self):
        return os.environ.get(env_vars.LOG_LEVEL) == 'debug'

    ### CRUD

    def to_hash(self):
        return self.config

    def update(self, contents):
        fp = open(self.config_file_path, 'w')
        yaml.dump(contents, fp, allow_unicode=True)
        fp.close()

    def reload_config(self, event):
        Logger.instance().info(f"{self.LOG_ID}.reload_config")
        self.__load_config()

    def to_json(self, pretty_print=False):
        output = None
        settings_dict = self.__dict__
        settings_dict['is_headless'] = self.is_headless()
        settings_dict['is_debug'] = self.is_debug()

        settings_dict['env_vars'] = {}
        environ_vars = settings_dict['env_vars']

        # https://stackoverflow.com/questions/11637293/iterate-over-object-attributes-in-python
        env_vars_fields = [a for a in dir(env_vars) if not a.startswith('__')]

        for env_var in env_vars_fields:
            environ_vars[env_var] = os.environ.get(env_vars.__dict__[env_var])

        if pretty_print:
            output = json.dumps(settings_dict, indent=4)
        else:
            output = json.dumps(settings_dict)

        return output

    ### Properties

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    @property
    def agent_url(self):
        if self._agent_url:
            return self._agent_url

        if os.environ.get(env_vars.AGENT_URL):
            return os.environ[env_vars.AGENT_URL]

        return self.config.get('agent_url')

    @agent_url.setter
    def agent_url(self, value):
        self._agent_url = value

    @property
    def proxy_url(self):
        if self._proxy_url:
            return self._proxy_url

        if os.environ.get(env_vars.AGENT_PROXY_URL):
            return os.environ[env_vars.AGENT_PROXY_URL]

        return self.config.get('proxy_url')

    @proxy_url.setter
    def proxy_url(self, value):
        self._proxy_url = value

    @property
    def api_url(self):
        if self._api_url:
            return self._api_url

        if os.environ.get(env_vars.API_URL):
            return os.environ[env_vars.API_URL]

        return self.config.get('api_url')

    @api_url.setter
    def api_url(self, value):
        self._api_url = value

    @property
    def api_key(self):
        if os.environ.get(env_vars.API_KEY):
            return os.environ[env_vars.API_KEY]

        return self.config.get('api_key')

    @property
    def mode(self):
        return self.config.get('mode')

    @property
    def active_mode(self):
        mode = self.mode

        if not mode:
            return None
        else:
            if os.environ.get(env_vars.AGENT_ACTIVE_MODE):
                return os.environ[env_vars.AGENT_ACTIVE_MODE]

            return mode.get('active')

    # Get active mode settings first from environment variables
    # If the env var is null, get the setting from the yaml file
    @property
    def active_mode_settings(self):
        mode = self.mode

        if not mode:
            return None

        active_mode = self.active_mode

        if not active_mode:
            return None

        if self.is_headless():
            # Get settings from yaml file, replace setting with env var if set
            active_mode_settings = mode.get(active_mode)

            enabled = os.environ.get(env_vars.AGENT_ENABLED)
            if enabled != None:
                active_mode_settings['enabled'] = not not enabled

            include_patterns = os.environ.get(env_vars.AGENT_INCLUDE_PATTERNS)
            if include_patterns != None:
                # Split the string based on commas, strip whitespace
                active_mode_settings['include_patterns'] = list(map(str.strip, include_patterns.split(',')))

            exclude_patterns = os.environ.get(env_vars.AGENT_EXCLUDE_PATTERNS)
            if exclude_patterns != None:
                # Split the string based on commas, strip whitespace
                active_mode_settings['exclude_patterns'] = list(map(str.strip, exclude_patterns.split(',')))

            policy = os.environ.get(env_vars.AGENT_POLICY)
            if policy != None:
                active_mode_settings['policy'] = policy

            proxy_service_url = os.environ.get(env_vars.AGENT_SERVICE_URL)
            if proxy_service_url != None:
                active_mode_settings['service_url'] = proxy_service_url

            proxy_project_key = os.environ.get(env_vars.AGENT_PROJECT_KEY)
            if proxy_project_key != None:
                active_mode_settings['project_key'] = proxy_project_key

            proxy_scenario_key = os.environ.get(env_vars.AGENT_SCENARIO_KEY)
            if proxy_scenario_key != None:
                active_mode_settings['scenario_key'] = proxy_scenario_key

            return active_mode_settings

        return mode.get(active_mode)

    ### Helpers

    def __load_config(self):
        with open(self.config_file_path, 'r') as stream:
            try:
                self.config = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                pass



#!/usr/bin/env python

import click
import distro
import json
import os
import pdb
import threading
import time

from .api import run as run_api
from .lib import env_vars
from .lib.ca_cert_installer import CACertInstaller
from .lib.cli.exec import run_command, run_command_with_proxy_export
from .lib.settings import Settings
from .proxy import run as run_proxy, get_proxy_url

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
@click.version_option()
@click.group(
    epilog="Run 'stoobly COMMAND --help' for more information on a command.",
    context_settings=CONTEXT_SETTINGS,
)
@click.pass_context
def main(ctx):
    pass

@click.group()
@click.pass_context
def config(ctx):
    pass

@click.group()
@click.pass_context
def ca_cert(ctx):
    pass

@main.command()
@click.option('--headless', is_flag=True, default=False, help='Disable starting UI.')
@click.option('--log-level', default='info', help='''
    Log level can be "debug", "info", "warning" or "error".
''')
@click.option('--mode', default="regular", help='''
    Mode can be "regular", "transparent", "socks5",
    "reverse:SPEC", or "upstream:SPEC". For reverse and
    upstream proxy modes, SPEC is host specification in
    the form of "http[s]://host[:port]".
'''
)
@click.option('--ssl-insecure', is_flag=True, default=False, help='Do not verify upstream server SSL/TLS certificates.')
@click.option('--proxy-host', default='0.0.0.0', help='Address to bind proxy to.')
@click.option('--proxy-port', default=8080, help='Proxy service port.')
@click.option('--ui-host', default='0.0.0.0', help='Address to bind UI to.')
@click.option('--ui-port', default=4200, help='UI service port.')
@click.option('--api-url', help='API URL.')
def run(**kwargs):
    if not os.getenv(env_vars.LOG_LEVEL):
        os.environ[env_vars.LOG_LEVEL] = kwargs['log_level']

    os.environ[env_vars.AGENT_PROXY_URL] = f"http://{kwargs['proxy_host']}:{kwargs['proxy_port']}"
    os.environ[env_vars.AGENT_URL] = f"http://{kwargs['ui_host']}:{kwargs['ui_port']}"

    if kwargs['api_url']:
        os.environ[env_vars.API_URL] = kwargs['api_url']

    if not kwargs['headless']:
        initialize_ui(kwargs)

    initialize_proxy(kwargs)

@main.command()
@click.option('--command', is_flag=True, default=False, help='Read commands from the command_string operand instead of from the standard input.')
@click.option('--shell', default='sh', help='Shell script to run interpret command(s).')
@click.argument('file_path')
def exec(**kwargs):
    is_command = kwargs['command']
    shell = kwargs['shell']
    file_path = kwargs['file_path']
    proxy_url = get_proxy_url()

    if not proxy_url:
        run_command(shell, file_path, is_command)
    else:
        run_command_with_proxy_export(shell, file_path, is_command, proxy_url)

@ca_cert.command()
def install(**kwargs):
    distro_name = distro.name(pretty=True)

    installer = CACertInstaller()

    # Ubuntu or other Debian based
    if distro.like() == 'debian':
        print(f"Installing CA certificate for {distro_name}...")
        installer.handle_debian()
    # MacOS
    elif distro.id() == 'darwin':
        installer.handle_darwin()
    # elif distro.id() == 'rhel':
    #     return
    else:
        print(f"{distro_name} is not supported yet for automatic CA cert installation.")

@ca_cert.command()
def uninstall():
    return

@config.command()
@click.option('--pretty-print', is_flag=True, default=False, help='Pretty print the json.')
@click.option('--save-to-file', is_flag=True, default=False, help='To save to a file or not.')
def dump(**kwargs):
    settings = Settings.instance()

    output = settings.to_json(pretty_print=kwargs['pretty_print'])

    if kwargs['save_to_file']:
        timestamp = str(int(time.time() * 1000))
        config_dump_file_name = f"stoobly_agent_config_dump_{timestamp}.json"

        with open(config_dump_file_name, 'w') as output_file:
            output_file.write(output)

        print(f"Config successfully dumped to {config_dump_file_name}")
    else:
        print(output)

# Attach subcommands to main
main.add_command(ca_cert)
main.add_command(config)

### Helpers

def initialize_ui(kwargs):
    ui_host = kwargs['ui_host']
    ui_port = kwargs['ui_port']

    print(f"UI server listening at http://{ui_host}:{ui_port}\n")

    thread = threading.Thread(target=run_api, args=(ui_host, ui_port))
    thread.start()

def initialize_proxy(kwargs):
    options = kwargs.copy()

    # Filter out non-mitmproxy options
    options['listen_host'] = options['proxy_host']
    options['listen_port'] = options['proxy_port']

    del options['headless']
    del options['log_level']
    del options['proxy_host']
    del options['proxy_port']
    del options['ui_host']
    del options['ui_port']
    del options['api_url']

    run_proxy(**options)



import os
import pdb

from mitmdump import DumpMaster, Options
from mitmproxy.optmanager import _Option

from .lib.root_dir import RootDir

PROXY_URL_FILENAME = 'proxy_url'
INTERCEPT_HANDLER_FILENAME = 'intercept_handler.py'

def run(**kwargs):
    __create_proxy_url_file(kwargs.get('listen_host'), kwargs.get('listen_port'))

    fixed_options = {
        'flow_detail': 1,
        'scripts': __get_intercept_handler_path(),
        'upstream_cert': False,
    }

    opts = Options(**{**kwargs, **fixed_options})
    __set_connection_strategy(opts, 'lazy')

    m = DumpMaster(opts)
    m.run()

def get_proxy_url():
    path = __proxy_url_abs_path()
    if not os.path.exists(path):
        return None

    with open(path) as f:
        return f.read()

def __get_intercept_handler_path():
    cwd = os.path.dirname(os.path.realpath(__file__))
    script = os.path.join(cwd, INTERCEPT_HANDLER_FILENAME)
    return script
##
#
# Equivalent of:
# mitmdump connection_strategy={strategy}
#
def __set_connection_strategy(opts, strategy):
    extra_options = {
        'dumper_filter': f"connection_strategy={strategy}",
        'readfile_filter': f"connection_strategy={strategy}",
        'save_stream_filter': f"connection_strategy={strategy}",
    }
    for k, v in extra_options.items():
        opts._options[k] = _Option(k, str, v, '', None)
    opts.update(**extra_options)

def __proxy_url_abs_path():
    tmp_dir_path = RootDir.instance().tmp_dir
    return os.path.join(tmp_dir_path, PROXY_URL_FILENAME)

def __create_proxy_url_file(host, port):
    file_path = __proxy_url_abs_path()
    tmp_dir_path = os.path.dirname(file_path)
    if not os.path.exists(tmp_dir_path):
        os.mkdir(tmp_dir_path)

    with open(file_path, 'w') as f:
        f.write(f"http://{host}:{port}")

def __remove_proxy_url_file():
    file_path = __proxy_url_abs_path()
    if os.path.exists(file_path):
        os.remove(file_path)



import base64
import pdb

from .response import  Response

class MitmproxyResponseAdapter(Response):

    def __init__(self, response):
        self.response = response
        self.content = response.raw_content

    @property
    def code(self):
        return self.response.status_code

    @property
    def headers(self):
        return self.response.headers

    def decode_body(self):
        # Decodes content (if Content-Encoding header is set)
        self.content = response.content

        # Update Content-Lenght header to decoded content length
        self.response.headers['content-length'] = str(len(self.content))

    @property
    def body(self):
        content = self.content

        if not content:
            return b''

        return content



class ProxyRequest:
    def __init__(self, request, upstream_url):
        self.request = request

        self.upstream_url = upstream_url

    def url(self):
        url = self.request.url

        return url.replace(self.request.base_url, self.upstream_url)



import json
import pdb
import urllib.parse

class RequestBodyParser:
    JSON = 'application/json'
    WWW_FORM_URLENCODED = 'application/x-www-form-urlencoded'

    ###
    #
    # @param request [ActionDispatch::Request]
    #
    # @return [Hash]
    #
    @classmethod
    def parse(cls, request):
        content = request.body

        if isinstance(content, bytes):
            content = content.decode('utf-8')

        params = {}

        content_type = request.content_type
        if not content_type:
            content_type = ''

        content_type = content_type.lower()

        if content_type == cls.JSON:
            params = cls.__parse_json(content)
        elif content_type == cls.WWW_FORM_URLENCODED:
            params = cls.__parse_www_form_urlencoded(content)

        return params

    @staticmethod
    def __parse_json(content):
        try:
            return json.loads(content)
        except:
            return {}

    @staticmethod
    def __parse_www_form_urlencoded(content):
        try:
            return urllib.parse.parse_qs(content)
        except:
            return {}



import os
from setuptools import find_packages, setup

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

public_files_relative_path = 'stoobly_agent/public'
public_files = package_files(public_files_relative_path)

setup(
    author='Michael Yen',
    author_email='michael@stoobly.com',
    description='Client agent for Stoobly',
    entry_points={
        'console_scripts': [
            'stoobly-agent=stoobly_agent.cli:main'
        ]
    },
    include_package_data=True,
    install_requires=[
        "click>=7.0.0,<8.0.0",
        "distro>=1.6.0,<1.7.0",
        "mergedeep>=1.3.0,<1.3.4",
        "mitmdump>=1.1.0,<=1.1.2",
        "mitmproxy>=6.0.0,<7.0.0",
        "pyyaml>=5.4,<=5.4.1",
        "requests>=2.25.0,<=2.25.1",
        "watchdog>=2.1.0,<=2.1.3",
    ],
    license='MIT',
    name='stoobly-agent',
    packages=find_packages(include=[
        'stoobly_agent', 'stoobly_agent.*',
    ]),
    package_data={
        'stoobly_agent': ['app/*' , 'config/*'] + public_files
    },
    #scripts=['bin/stoobly-agent'],
    url='https://github.com/Stoobly/stoobly-agent',
    version='0.5.2',
)




import hashlib
import pdb

from urllib.parse import parse_qs

from .logger import Logger
from .request_body_parser import RequestBodyParser

class HashedRequestDecorator:

    LOG_ID = 'lib.hashed_request_decorator'

    COMPONENT_TYPES = {
        'HEADER': 1,
        'PATH_SEGMENT': 2,
        'QUERY_PARAM': 3,
        'BODY_PARAM': 4,
        'RESPONSE': 5
    }

    def __init__(self, request):
        self.request = request

        self.ignored_headers = {}
        self.ignored_query_params = {}
        self.ignored_body_params = {}

    ###
    #
    # self.param ignored_components [Array<Hash>] e.g. [{"name":"script-name","type":1}]
    #
    def with_ignored_components(self, ignored_components):
        for ignored_component in ignored_components:
            component_name = ignored_component['name']
            component_type = ignored_component['type']

            if component_type == self.COMPONENT_TYPES['HEADER']:
                self.ignored_headers[component_name] = True
            elif component_type == self.COMPONENT_TYPES['QUERY_PARAM']:
                self.ignored_query_params[component_name] = True
            elif component_type == self.COMPONENT_TYPES['BODY_PARAM']:
                self.ignored_body_params[component_name] = True

        return self

    def headers_hash(self, with_ignored = False):
        headers = self.request.headers

        serialized_params = self.__serialize_params(headers, {} if with_ignored else self.ignored_headers)

        return self.__hash_serialized_params(serialized_params)

    def query_params_hash_with_ignored(self):
        return self.query_params_hash(True)

    def query_params_hash(self, with_ignored = False):
        query_params = self.request.query

        params = self.__deflatten_multi_dict(query_params)

        serialized_params = self.__serialize_params(params, {} if with_ignored else self.ignored_query_params)

        return self.__hash_serialized_params(serialized_params)

    def body_params_hash_with_ignored(self):
        return self.body_params_hash(True)

    def body_params_hash(self, with_ignored = False):
        params = RequestBodyParser.parse(self.request)

        serialized_params = self.__serialize_params(params, {} if with_ignored else self.ignored_body_params)

        return self.__hash_serialized_params(serialized_params)

    def body_text_hash(self):
        text = self.request.body

        if len(text) == 0:
            return text

        return self.__hash(text)

    def __hash_serialized_params(self, serialized_params):
        if len(serialized_params) == 0:
            return ''

        serialized_params.sort()
        return self.__hash('.'.join(serialized_params))

    def __hash(self, text):
        return hashlib.md5(text.encode('utf-8')).hexdigest()

    def __serialize_param(self, key, val):
        return f"{key}.{str(val)}".encode('utf-8')

    def __serialize_params(self, params, ignored_params = {}):
        serialized_params = []

        for key, value in params.items():
            if key in ignored_params:
                continue

            if isinstance(value, list):
                for param in value:
                    param_hash = hashlib.md5(self.__serialize_param(key, param)).hexdigest()

                    Logger.instance().debug(f"{self.LOG_ID}.serialized_query_params_hash:{key} -> {param} ({param_hash})")

                    serialized_params.append(param_hash)
            else:
                param_hash = hashlib.md5(self.__serialize_param(key, value)).hexdigest()

                Logger.instance().debug(f"{self.LOG_ID}.serialized_query_params_hash:{key} -> {value} ({param_hash})")

                serialized_params.append(param_hash)

        return serialized_params

    def __deflatten_multi_dict(self, multi_dict):
        params = {}
        for name, value in multi_dict.items(multi=True):
            if name not in params:
                params[name] = value
            else:
                if not isinstance(params[name], list):
                   params[name] = [params[name]]

                params[name].append(value)
        return params



import os
import subprocess

class CACertInstaller():

    def __init__(self):
        home_dir = os.path.expanduser('~')
        self.mitmproxy_certs_dir = os.path.join(home_dir, '.mitmproxy')
        self.pem_file_name = 'mitmproxy-ca-cert.pem'
        self.crt_file_name = 'mitmproxy-ca-cert.crt'

    # https://askubuntu.com/a/94861
    def handle_debian(self):
        extra_ca_certs_dir = '/usr/local/share/ca-certificates/extra'
        mitm_pem_absolute_path = os.path.join(self.mitmproxy_certs_dir, self.pem_file_name)
        mitm_crt_absolute_path = os.path.join(self.mitmproxy_certs_dir, self.crt_file_name)
        extra_crt_absolute_path = os.path.join(extra_ca_certs_dir, self.crt_file_name)

        subprocess.run(f"sudo mkdir -p {extra_ca_certs_dir}".split(), check=True)
        subprocess.run(f"openssl x509 -in {mitm_pem_absolute_path} -inform PEM -out {mitm_crt_absolute_path}".split(), check=True)
        subprocess.run(f"sudo cp {mitm_crt_absolute_path} {extra_crt_absolute_path}".split(), check=True)
        subprocess.run('sudo update-ca-certificates'.split(), check=True)

    # https://www.dssw.co.uk/reference/security.html
    def handle_darwin(self):
        system_keychain_path = '/Library/Keychains/System.keychain'
        mitm_crt_absolute_path = os.path.join(self.mitmproxy_certs_dir, self.crt_file_name)

        subprocess.run(f"sudo security add-trusted-cert -d -p ssl -p basic -k {system_keychain_path} {mitm_crt_absolute_path}".split(), check=True)



# Modeled after Net::HTTP::Response

class Response:
    def __init__(self):
        self._code = 0
        self._headers = {}
        self._body = ''

    @property
    def code(self):
        return self._code

    @property
    def header(self):
        return self._headers

    @property
    def body(self):
        return self._body



# Rails request object
# https://guides.rubyonrails.org/action_controller_overview.html#the-request-object
class Request:
    def __init__(self):
        self._method = ''
        self._url = ''
        self._path = ''
        self._base_url = ''
        self._headers = {}
        self._body = ''
        self._query = {}
        self._content_type = ''

    @property
    def method(self):
        return self._method

    @property
    def url(self):
        return self._url

    @property
    def base_url(self):
        return self._base_url

    @property
    def path(self):
        return self._path

    @property
    def headers(self):
        return self._headers

    @property
    def body(self):
        return self._body

    @property
    def query(self):
        return self._query

    @property
    def content_type(self):
        return self._content_type




import pdb

from urllib.parse import urlparse

from .request import Request

class MitmproxyRequestAdapter(Request):
    def __init__(self, request):
        self.request = request

    @property
    def url(self):
        return self.request.url

    @property
    def path(self):
        uri = urlparse(self.request.path)
        return uri.path

    @property
    def base_url(self):
        return f"{self.request.scheme}://{self.request.host}:{self.request.port}"

    @property
    def method(self):
        return self.request.method

    @property
    def headers(self):
        return self.request.headers

    @property
    def body(self):
        content = self.request.content

        try:
            if isinstance(content, bytes):
                content = content.decode('utf-8')
        except:
            return ''.join(map(chr, content))

        return content

    @property
    def query(self):
        return self.request.query

    @property
    def content_type(self):
        for key, value in self.headers.items():
            if key.lower() == 'content-type':
                return value
        return ''

    @property
    def host(self):
        return self.request.host


    @property
    def port(self):
        return self.request.port




import hashlib
import time
import pdb

class RequestString:
    ENCODING = 'utf-8'
    REQUEST_TYPE = 1
    CLRF = "\r\n"

    def __init__(self, proxy_request):
        self.request = proxy_request.request
        self.proxy_request = proxy_request

        self.lines = []

        self.request_line()
        self.headers()
        self.body()

        self.request_id = self.generate_request_id()
        self.control()

    def get(self):
        return self.CLRF.join(self.lines).encode(self.ENCODING)

    def control(self):
        current_time = self.current_time()
        self.lines.insert(0, "{} {} {}".format(self.REQUEST_TYPE, self.request_id, current_time))


    def request_line(self):
        self.lines.append("{} {} HTTP/1.1".format(self.request.method, self.proxy_request.url()))

    def headers(self):
        _headers = self.request.headers

        for name, val in _headers.items():
            line = ' '.join(["{}:".format(self.to_header_case(name)), val])
            self.lines.append(line)

    def body(self):
        self.lines.append("{}{}".format(self.CLRF, self.request.body))

    def to_header_case(self, header):
        toks = header.split('_')

        for index, tok in enumerate(toks):
            toks[index] = tok.lower().capitalize()

        return "-".join(toks)

    def generate_request_id(self):
        joined_lines = self.CLRF.join(self.lines)
        return hashlib.md5(joined_lines.encode(self.ENCODING)).hexdigest()

    def current_time(self):
        now = time.time()
        current_time = round(now * (pow(10, 9)))

        return current_time
